<?php

namespace Drupal\xray_audit\Plugin\xray_audit\tasks\QueryGroup;

use Drupal\Core\Database\Query\Insert;

/**
 * Plugin implementation of queries_data_node.
 *
 * @XrayAuditTaskPlugin (
 *   id = "queries_data_paragraphs",
 *   label = @Translation("Data about paragraphs"),
 *   description = @Translation("Reports on the Paragraphs of this site."),
 *   group = "queries_data",
 *   sort = 4,
 *   operations = {
 *      "paragraphs_count" = {
 *          "label" = "Paragraphs types and count",
 *          "description" = "Paragraphs types and count (default language)."
 *       },
 *      "paragraphs_count_by_hierarchy" = {
 *          "label" = "Paragraphs count grouped by type ordered hierarchically",
 *          "description" = "Paragraphs count grouped by type ordered
 *            hierarchically (default language)."
 *       },
 *      "paragraphs_count_by_language" = {
 *          "label" = "Paragraphs count grouped by language",
 *          "description" = "Paragraphs count grouped by language."
 *       },
 *    },
 *   dependencies = {"paragraphs"},
 *   install = "paragraphsIntallActions",
 *   uninstall = "paragraphsUnintallActions"
 * )
 */
class XrayAuditQueryTaskParagraphsPlugin extends XrayAuditQueryTaskPluginBase {

  const XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME = 'xra_paragraphs_hierarchy_tmp';

  /**
   * {@inheritdoc}
   */
  public function getDataOperationResult(string $operation = NULL) {
    switch ($operation) {
      case 'paragraphs_count':
        return $this->paragraphsTypes();

      case 'paragraphs_count_by_hierarchy':
        return $this->paragraphsTypesHierarchy();

      case 'paragraphs_count_by_language':
        return $this->paragraphsItemsCountByLanguage();
    }
    return [];
  }

  /**
   * Get data of Paragraph types.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsTypes() {
    $headerTable = [
      $this->t('id'),
      $this->t('label'),
      $this->t('count'),
    ];
    $resultTable = [];

    $storage_type = $this->entityTypeManager->getStorage('paragraphs_type');
    $storage = $this->entityTypeManager->getStorage('paragraph');
    $language_by_default = $this->languageManager->getDefaultLanguage()->getId();
    $alias_count = 'count';

    $result = $storage
      ->getAggregateQuery()
      ->accessCheck(FALSE)
      ->aggregate('id', 'count', $language_by_default, $alias_count)
      ->groupBy('type')
      ->execute();

    // Prepare results.
    $result_processed = [];
    foreach ($result as $row) {
      $result_processed[$row['type']] = $row['count'];
    }
    $total = 0;

    $paragraph_types = $storage_type->loadMultiple();
    foreach ($paragraph_types as $key => $paragraph_type) {
      $resultTable[] = [
        $key,
        $paragraph_type->label(),
        ($result_processed[$key] ?? 0),
      ];
      $total = $total + ($result_processed[$key] ?? 0);
    }

    $resultTable[] = [
      '',
      $this->t('Total'),
      $total,
    ];

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];
  }

  /**
   * Get data of Paragraph types ordered hierarchically.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsTypesHierarchy(): array {
    $this->paragraphsHierarchyCreateTemporaryTable();

    $headerTable = [
      $this->t('Level'),
      $this->t('Parent'),
      '',
      $this->t('Count'),
    ];
    $language_by_default = $this->languageManager->getDefaultLanguage()->getId();

    $query = $this->database->select(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME, 'tmp')
      ->fields('tmp', ['level', 'parent_bundle', 'type']);
    $query->addExpression('COUNT(tmp.id)', 'count');
    $results = $query->condition('tmp.langcode', $language_by_default)
      ->groupBy('tmp.hierarchy_id')
      ->groupBy('tmp.parent_bundle')
      ->groupBy('tmp.level')
      ->groupBy('tmp.type')
      ->orderBy('tmp.hierarchy_id')
      ->execute()
      ->fetchAll();

    // Prepare results.
    $resultTable = [];
    $paragraph_types = $this->entityTypeManager->getStorage('paragraphs_type')->loadMultiple();
    $total = 0;

    foreach ($results as $result) {
      $result = (array) $result;
      $indent_string = $result['level'] != '1' ? str_repeat(' ', $result['level'] - 1) . '- ' : NULL;
      $parent_label = $result['level'] == '1' ? $result['parent_bundle'] : NULL;
      $resultTable[] = [
        $result['level'],
        $parent_label,
        $indent_string . $paragraph_types[$result['type']]->label() . ' (' . $result['type'] . ')',
        $result['count'],
      ];
      $total = $total + $result['count'];
    }

    $resultTable[] = [
      '',
      '',
      $this->t('Total'),
      $total,
    ];

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];
  }

  /**
   * Paragraphs items count by language.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsItemsCountByLanguage() {

    $headerTable = [
      $this->t('label'),
      $this->t('type'),
      $this->t('langcode'),
      $this->t('total'),
    ];
    $resultTable = [];

    $storage_type = $this->entityTypeManager->getStorage('paragraphs_type');
    $storage = $this->entityTypeManager->getStorage('paragraph');
    $alias_count = 'count';

    // Get label of content types.
    $label = [];
    $types = $storage_type->loadMultiple();
    foreach ($types as $key => $type) {
      $label[$key] = $type->label();
    }

    $query = $storage->getAggregateQuery();
    $result = $query->accessCheck(FALSE)
      ->aggregate('id', 'COUNT', NULL, $alias_count)
      ->groupBy('langcode')
      ->groupBy('type')
      ->execute();

    foreach ($result as $row) {
      $resultTable[] = [
        'label' => $label[$row['type']],
        'id' => $row['type'],
        'langcode' => $row['langcode'],
        'total' => $row['count'] ?? 0,
      ];
    }

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];

  }

  /**
   * Create/update the paragraphs Hierarchy temporally table.
   */
  protected function paragraphsHierarchyCreateTemporaryTable(): void {
    $state_service = \Drupal::service('state');
    $last_execution = $state_service->get('xray_audit.paragraps_hierarchy_temporary_table_creation', 0);

    if ($last_execution > strtotime("-1 day")) {
      return;
    }

    $this->database->truncate(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)->execute();

    $results = $this->database->select('paragraphs_item_field_data')
      ->fields('paragraphs_item_field_data', [
        'id',
        'parent_id',
        'type',
        'langcode',
        'parent_type',
      ])
      ->condition('parent_type', 'paragraph', '<>')
      ->execute()
      ->fetchAll();

    $query = $this->database->insert(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)
      ->fields([
        'hierarchy_id',
        'id',
        'parent_id',
        'type',
        'parent_type',
        'parent_bundle',
        'langcode',
        'level',
      ]);

    foreach ($results as $result) {
      $result = (array) $result;
      $result['hierarchy_id'] = $result['parent_type'] . "-" . $result['type'];
      $result['parent_bundle'] = $result['parent_type'];
      $result['level'] = 1;
      $query->values($result);
      $this->paragraphsHierarchyProcessLevel($result, $query);
    }

    $query->execute();

    $state_service->set('xray_audit.paragraps_hierarchy_temporary_table_creation', time());
  }

  /**
   * Process a level of the paragraphs Hierarchy temporally table.
   *
   * @param array $parent
   *   Parent result.
   * @param \Drupal\Core\Database\Query\Insert $query
   *   Drupal Query object.
   */
  protected function paragraphsHierarchyProcessLevel(array $parent, Insert $query): void {
    $results = $this->database->select('paragraphs_item_field_data')
      ->fields('paragraphs_item_field_data', [
        'id',
        'parent_id',
        'type',
        'langcode',
        'parent_type',
      ])
      ->condition('parent_id', $parent['id'])
      ->condition('parent_type', 'paragraph')
      ->execute()
      ->fetchAll();

    foreach ($results as $result) {
      $result = (array) $result;
      $result['hierarchy_id'] = $parent['hierarchy_id'] . "-" . $result['type'];
      $result['parent_bundle'] = $parent['type'];
      $result['level'] = $parent['level'] + 1;
      $query->values($result);

      if ($result['level'] <= 10) {
        $this->paragraphsHierarchyProcessLevel($result, $query);
      }
    }
  }

  /**
   * Get the table configuration for the paragraphs Hierarchy temporally table.
   *
   * @return array
   *   Table configuration array.
   */
  protected function paragraphsHierarchyGetTableConfiguration(): array {
    return [
      'description' => 'Temporary table to get paragraph usage with hierarchy',
      'fields' => [
        'hierarchy_id' => [
          'type' => 'varchar',
          'length' => 255,
        ],
        'id' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '0',
          'unsigned' => TRUE,
        ],
        'parent_id' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '0',
          'unsigned' => TRUE,
        ],
        'type' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'parent_type' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'parent_bundle' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'langcode' => [
          'type' => 'varchar',
          'length' => 10,
        ],
        'level' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '1',
        ],
      ],
    ];
  }

  /**
   * Create the requirements for the task.
   */
  public function paragraphsIntallActions(): void {
    if (!$this->database->schema()->tableExists(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)) {
      $this->database->schema()->createTable(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME, $this->paragraphsHierarchyGetTableConfiguration());
    }
  }

  /**
   * Uninstall the requirements for the task.
   */
  public function paragraphsUnintallActions(): void {
    $this->database->schema()->dropTable(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME);
  }

}
